/*
 *  dumpfile.c
 *
 *  Copyright (C) 2005, 2006  FUJITSU LIMITED
 *
 */
/* $Id: dumpfile.c,v 1.8 2006/12/07 23:38:57 akira Exp $ */

#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include "savecore.h"

DumpFile *open_dump_file(char *dir)
{
	DumpFile *dump_file;
	char *filename;
	int fd;

	if ((dump_file = calloc(1, sizeof(DumpFile))) == NULL) {
		return NULL;
	}

	dump_file->dir = dir;

	if (asprintf(&filename, "%s%s", dir, INCOMPLETE_DUMP_FILE) < 0) {
		return NULL;
	}

	fd = open(filename, O_WRONLY|O_TRUNC|O_CREAT, S_IRUSR|S_IWUSR);
	if (fd < 0) {
		free(dump_file);
		free(filename);
		return NULL;
	}

	dump_file->fd	= fd;
	dump_file->name	= filename;

	return dump_file;
}

int rename_dump_file(DumpFile *dump_file)
{
	char *old_file;
	char *new_file;
	int ret;

	if (asprintf(&old_file, "%s%s", dump_file->dir, INCOMPLETE_DUMP_FILE)
			< 0) {
		return FAILED;
	}
	if (asprintf(&new_file, "%s%s", dump_file->dir, COMPLETE_DUMP_FILE)
			< 0) {
		return FAILED;
	}

	ret = rename(old_file, new_file);
	if ((dump_file->name = strdup(new_file)) == NULL) {
		return FAILED;
	}

	free(old_file);
	free(new_file);

	return ret;
}

int read_page_desc(DumpFile *dump_file, page_desc_t *desc)
{
	if (dump_seek(dump_file, dump_file->page_desc_offset) < 0)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (read(dump_file->fd, desc, sizeof(*desc)) != sizeof(*desc))
		return READ_ERROR_IN_DUMP_FILE;

	dump_file->page_desc_offset += sizeof(*desc);

	return COMPLETE;
}

int write_page_desc(DumpFile *dump_file, page_desc_t *desc)
{
	if (dump_seek(dump_file, dump_file->page_desc_offset) < 0)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (write(dump_file->fd, desc, sizeof(*desc)) != sizeof(*desc))
		return WRITE_ERROR_IN_DUMP_FILE;

	dump_file->page_desc_offset += sizeof(*desc);

	return COMPLETE;
}

int write_page_data(DumpFile *dump_file, const void *data, page_desc_t *desc)
{
	if (dump_seek(dump_file, desc->offset) < 0)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (write(dump_file->fd, data, desc->size) != desc->size)
		return WRITE_ERROR_IN_DUMP_FILE;

	dump_file->data_offset = desc->offset + desc->size;

	return COMPLETE;
}

int read_page_data(DumpFile *dump_file, void *data, page_desc_t *desc)
{
	if (dump_seek(dump_file, desc->offset) < 0)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (read(dump_file->fd, data, desc->size) != desc->size)
		return READ_ERROR_IN_DUMP_FILE;

	dump_file->data_offset += desc->size;

	return COMPLETE;
}

int dump_read(DumpFile *dump_file, void *data, uint32_t size)
{
	if (dump_seek(dump_file, dump_file->data_offset) < 0)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (read(dump_file->fd, data, size) != size)
		return READ_ERROR_IN_DUMP_FILE;

	dump_file->data_offset
		+= roundup(size, dump_file->device->header->block_size);

	return COMPLETE;
}

int dump_write(DumpFile *dump_file, const void *data, uint32_t size)
{
	if (dump_seek(dump_file, dump_file->data_offset) < 0)
		return SEEK_ERROR_IN_DUMP_DEVICE;

	if (write(dump_file->fd, data, size) != size)
		return WRITE_ERROR_IN_DUMP_FILE;

	dump_file->data_offset
		+= roundup(size, dump_file->device->header->block_size);

	return COMPLETE;
}

void *alloc_blocks(DumpFile *dump_file, int block_nr)
{
	return calloc(1, dump_file->device->header->block_size * block_nr);
}

int read_blocks(DumpFile *dump_file, void *data, int block_nr)
{
	return dump_read(dump_file, data,
			 block_nr * dump_file->device->header->block_size);
}

int write_blocks(DumpFile *dump_file, const void *data, int block_nr)
{
	return dump_write(dump_file, data,
			  block_nr * dump_file->device->header->block_size);
}

void set_data_offset(DumpFile *dump_file)
{
	dump_file->data_offset = dump_file->device->header->block_size
				* (1 + dump_file->device->header->sub_hdr_size);
}

void calc_data_offset(DumpFile *dump_file)
{
	int i;
	unsigned long num_pages = 0;

	for (i = 0; i <= dump_file->device->header->max_mapnr; i++)
		if (page_is_dumpable(i, dump_file->device))
			num_pages++;
	dump_file->data_offset += num_pages * sizeof(page_desc_t);
}
