/*
 *  dumpfilter.c
 *
 *  Copyright (C) 2006  FUJITSU LIMITED
 *
 */
/* $Id: dumpfilter.c,v 1.8 2006/12/07 23:38:57 akira Exp $ */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>

#include "savecore.h"
#include "dumpfilter.h"

#define CMD	"dumpfilter"

static int dump_level = DEFAULT_DUMP_LEVEL;
static int skipped_page_cnt = 0;
static int saved_page_cnt = 0;
static bool prog_flg = false;	/* becomes true if option '-p' specified. */
static const int symbol = '#';	/* for progress bar */

static DumpFile *open_vmcore(const char *name, int flag)
{
	DumpFile *vmcore;

	if ((vmcore = calloc(1, sizeof(*vmcore))) == NULL) {
		return NULL;
	}

	if ((vmcore->device = calloc(1, sizeof(DumpDevice))) == NULL) {
		goto bad;
	}

	switch (flag) {
	case NEW:
		if ((vmcore->fd = open(name,
			O_WRONLY|O_TRUNC|O_CREAT, S_IRUSR|S_IWUSR)) < 0) {
			goto bad;
		}
		break;

	case EXISTING:
	default:
		if ((vmcore->fd = open(name, O_RDONLY)) < 0) {
			goto bad;
		}
		break;
	}

	vmcore->dir = "";
	vmcore->name = name;

	return vmcore;
bad:
	free(vmcore->device);
	free(vmcore);

	return NULL;
}

static inline void show_header_info(struct disk_dump_header *hdr, bool verbose)
{
	Info("header info\n");
	Info("  total device blocks: %d\n", hdr->device_blocks);
	Info("  bitmap length      : %d\n", hdr->bitmap_blocks);
	Info("  max map number     : %d\n", hdr->max_mapnr);
	Info("  total RAM blocks   : %d\n", hdr->total_ram_blocks);
	Info("  written blocks     : %d\n", hdr->written_blocks);
}

static inline void show_dump_info(struct disk_dump_header *hdr, bool verbose)
{
	const char *timeofdeath = get_date_time_str(hdr->timestamp.tv_sec);
	const char *current     = get_date_time_str(CURRENT_DATE_TIME);

	Info("dump info\n");
	Info("  kernel             : %s\n", kernel_path(hdr->utsname.release));
	Info("  node name          : %s\n", hdr->utsname.nodename);
	Info("  release            : %s\n", hdr->utsname.release);
	Info("  machine            : %s\n", hdr->utsname.machine);
	Info("  dumped date        : %s\n", timeofdeath);
	Info("  current date       : %s\n", current);
	Info("  passed days        : %d\n",
			diff_days(CURRENT_DATE_TIME, hdr->timestamp.tv_sec));
}

static void *read_alloc(int fd, size_t len)
{
	void *data;

	if ((data = calloc(1, len)) == NULL) {
		return NULL;
	}

	if (read(fd, data, len) < len) {
	 	free(data);
	 	return NULL;
	}

	return data;
}

static int check_vmcore_header(DumpFile *src)
{
	struct disk_dump_header *header;
	struct disk_dump_sub_header *sub_header;
	const int block_size = (int)sysconf(_SC_PAGE_SIZE);

	if (block_size < 0) {
		return BLOCK_SIZE_INVALID;
	}

	/* read vmcore header */
	if ((header = read_alloc(src->fd, block_size)) == NULL) {
		return READ_ERROR_IN_DUMP_HEADER;
	}

	if (!dump_signature_is_good(header->signature)) {
		free(header);
		return NO_PANIC_DUMP_HEADER;
	}
	if (header->block_size != block_size) {
		free(header);
		return BLOCK_SIZE_NOT_MATCH;
	}

	src->device->header = header;

	/* read vmcore sub header */
	if ((sub_header = read_alloc(src->fd, block_size)) == NULL) {
		free(header);
		return READ_ERROR_IN_DUMP_SUB_HEADER;
	}

	src->device->sub_header = sub_header;

	src->device->bitmap_len = block_size * header->bitmap_blocks;

	return COMPLETE;
}

static int write_vmcore_header(DumpFile *dst, DumpFile *src)
{
	/* save original data_offset */
	const off_t data_offset = dst->data_offset;

	/* write vmcore header */
	dst->data_offset = 0;
	if (write_blocks(dst, src->device->header, 1) != 0) {
		return WRITE_ERROR_IN_DUMP_HEADER;
	}

	/* write vmcore sub header */
	if (write_blocks(dst, src->device->sub_header, 1) != 0) {
		return WRITE_ERROR_IN_DUMP_SUB_HEADER;
	}

	dst->data_offset = data_offset;

	return COMPLETE;
}

static int read_vmcore_bitmap(DumpFile *src)
{
	if ((src->device->bitmap
		= alloc_blocks(src, src->device->header->bitmap_blocks))
			== NULL) {
		return UNABLE_TO_ALLOCATE;
	}

	set_data_offset(src);

	if (read_blocks(src, src->device->bitmap,
			src->device->header->bitmap_blocks) != COMPLETE) {
		return READ_ERROR_IN_MEMORY_BITMAP;
	}

	src->device->dumpable_bitmap = get_dumpable_bitmap(src->device);

	src->page_desc_offset = src->data_offset;

	calc_data_offset(src);

	return COMPLETE;
}

static int dup_vmcore(DumpFile *dst, DumpFile *src)
{
	dst->device->header		= src->device->header;
	dst->device->sub_header		= src->device->sub_header;
	dst->device->sub_header_len	= src->device->sub_header_len;
	dst->device->bitmap_len		= src->device->bitmap_len;

	dst->data_offset		= src->data_offset;
	dst->page_desc_offset		= src->page_desc_offset;

	if ((dst->device->bitmap
		= alloc_blocks(dst, dst->device->header->bitmap_blocks))
			== NULL) {
		return FAILED;
	}
	memcpy(dst->device->bitmap, src->device->bitmap,
			dst->device->bitmap_len);

	dst->device->dumpable_bitmap = get_dumpable_bitmap(dst->device);

	return SUCCEEDED;
}

static int write_vmcore_bitmap(DumpFile *dst)
{
	set_data_offset(dst);
	if (write_blocks(dst, dst->device->bitmap,
			dst->device->header->bitmap_blocks) != COMPLETE) {
		return FAILED;
	}

	return SUCCEEDED;
}

static inline int PageAnon(page_desc_t *pd)
{
	return pd->flags & DUMP_DH_MAPPING_ANON;
}

/*
 * determine whether to save the page or not.
 *   true:  page is saved
 *   false: page is not saved
 */
static bool page_saved(page_desc_t *pd)
{
	if (!dump_level) {
		return true;
	}
 	if ((dump_level & DUMP_EXCLUDE_CACHE) &&
	    (dump_level & DUMP_SAVE_PRIVATE) && PagePrivate(pd)) {
		return true;
	}
	if ((dump_level & DUMP_EXCLUDE_CACHE) && !PageAnon(pd) &&
	    (PageLRU(pd) || PageSwapCache(pd))) {
		return false;
	}
	if ((dump_level & DUMP_EXCLUDE_ANON) && PageAnon(pd)) {
		return false;
	}

	return true;
}

static int write_vmcore_pages(DumpFile *dst, DumpFile *src)
{
	int i;
	char *data;
	page_desc_t pd;

	if ((data = alloc_blocks(dst, 1)) == NULL) {
		return UNABLE_TO_ALLOCATE;
	}

	for (i = 0; i <= src->device->header->max_mapnr; i++) {
		if (prog_flg) {
			progress_bar(dst->name, symbol, i,
					src->device->header->max_mapnr);
		}

		if (!page_is_dumpable(i, src->device)) {
			continue;
		}

		if (read_page_desc(src, &pd) != COMPLETE) {
			free(data);
			return READ_ERROR_IN_PAGE_DESC;
		}

		if (!page_saved(&pd)) {
			page_dumpable_clear(i, dst->device);
			skipped_page_cnt++;
			continue;
		}

		saved_page_cnt++;

		if (read_page_data(src, data, &pd) != COMPLETE) {
			free(data);
			return READ_ERROR_IN_PAGE_DATA;
		}

		pd.offset = dst->data_offset;

		if (write_page_desc(dst, &pd) != COMPLETE) {
			free(data);
			return WRITE_ERROR_IN_PAGE_DESC;
		}

		if (write_page_data(dst, data, &pd) != COMPLETE) {
			free(data);
			return WRITE_ERROR_IN_PAGE_DATA;
		}
	}
	if (prog_flg) {
		progress_bar_post(dst->name, symbol, i,
				src->device->header->max_mapnr);
	}

	return COMPLETE;
}

static inline void get_dump_level(char *val)
{
	dump_level = strtol(val, NULL, 0);
}

int main(int argc, char *argv[])
{
	int opt;
	bool frc_flg = false, verbose = false;
	const char *input = NULL, *output = NULL;
	DumpFile *vmcore_i, *vmcore_o;

	while ((opt = getopt(argc, argv, "fl:pvV")) != -1) {
		switch (opt) {
		case 'f':
			frc_flg = true;
			break;
		case 'l':
			get_dump_level(optarg);
			break;
		case 'p':
			prog_flg = true;
			break;
		case 'v':
			verbose = true;
			break;
		case 'V':
			version(CMD);
			break;
		default:
			usage(CMD);
			break;
		}
	}

	if (optind + 2 != argc) {
		usage(CMD);
	}

	input = argv[optind];		/* get the source vmcore name */
	output = argv[optind + 1];	/* get the target vmcore name */

	if (!does_exist(input)) {
		Err("%s does not exist\n", input);
		exit(1);
	}

	if (does_exist(output)) {
		if (!frc_flg) {
			Err("%s already exists\n", output);
			exit(1);
		}
		if (same_file(input, output)) {
			Err("source and target are identical\n");
			exit(1);
		}
	}

	Info("filtering\n");
	Info("  dump level         : %d\n", dump_level);
	if (frc_flg) {
		Info("  note               : %s will be overwritten\n", output);
	}

	if ((vmcore_i = open_vmcore(input, EXISTING)) == NULL) {
		Err("unable to open %s\n", input);
		exit(1);
	}
	Info("input file\n");
	Info("  file name          : %s\n", vmcore_i->name);

	switch (check_vmcore_header(vmcore_i)) {
	case COMPLETE:
		break;
	case NO_PANIC_DUMP_HEADER:
		Err("%s does not have panic dump header.\n", vmcore_i->name);
		exit(1);
	case READ_ERROR_IN_DUMP_HEADER:
		Err("unable to read the dump header in %s\n", vmcore_i->name);
		exit(1);
	case READ_ERROR_IN_DUMP_SUB_HEADER:
		Err("unable to read the dump sub header in %s\n",
								vmcore_i->name);
		exit(1);
	case BLOCK_SIZE_INVALID:
		Err("block size is invalid.\n");
		exit(1);
	case BLOCK_SIZE_NOT_MATCH:
		Err("block size of %s does not match with system page size.\n",
								vmcore_i->name);
		exit(1);
	default:
		Err("unknown error (checking the dump header in %s)\n",
								vmcore_i->name);
		exit(1);
	}
	Info("  vmcore header      : checked\n");

	switch (read_vmcore_bitmap(vmcore_i)) {
	case COMPLETE:
		break;
	case UNABLE_TO_ALLOCATE:
		Err("unable to allocate memory to read the vmcore bitmap\n");
		exit(1);
	case READ_ERROR_IN_MEMORY_BITMAP:
		Err("unable to read bitmap in %s\n", vmcore_i->name);
		exit(1);
	default:
		Err("unknown error (reading the vmcore bitmap in %s)\n",
								vmcore_i->name);
		exit(1);
	}
	Info("  vmcore bitmap      : read\n");

	show_header_info(vmcore_i->device->header, verbose);
	show_dump_info(vmcore_i->device->header, verbose);

	if ((vmcore_o = open_vmcore(output, NEW)) == NULL) {
		Err("unable to open %s\n", output);
		exit(1);
	}
	Info("output file\n");
	Info("  file name          : %s\n", vmcore_o->name);

	if (dup_vmcore(vmcore_o, vmcore_i) == FAILED) {
		Err("unable to duplicate vmcore\n");
		goto err;
	}

	switch (write_vmcore_header(vmcore_o, vmcore_i)) {
	case COMPLETE:
		break;
	case WRITE_ERROR_IN_DUMP_HEADER:
		Err("unable to write data to header.\n");
		goto err;
	case WRITE_ERROR_IN_DUMP_SUB_HEADER:
		Err("unable to write data to sub header.\n");
		goto err;
	default:
		Err("unknown error (writing to the dump header)\n");
		goto err;
	}
	Info("  vmcore header      : written\n");

	if (!prog_flg) {
		Info("  vmcore pages       : ");
	}
	switch (write_vmcore_pages(vmcore_o, vmcore_i)) {
	case COMPLETE:
		break;
	case UNABLE_TO_ALLOCATE:
		Err("unable to allocate memory for bitmap\n");
		goto err;
	case READ_ERROR_IN_PAGE_DESC:
		Err("unable to read page descriptor\n");
		goto err;
	case READ_ERROR_IN_PAGE_DATA:
		Err("unable to read page data\n");
		goto err;
	case WRITE_ERROR_IN_PAGE_DESC:
		Err("unable to write page descriptor\n");
		goto err;
	case WRITE_ERROR_IN_PAGE_DATA:
		Err("unable to write page data\n");
		goto err;
	default:
		Err("unknown error (writing to the vmcore pages)\n");
		goto err;
	}
	if (!prog_flg) {
		Info("written\n");
	} else {
		Info("  vmcore pages       : written\n");
	}

	if (write_vmcore_bitmap(vmcore_o) == FAILED) {
		Err("unable to write vmcore bitmap to %s\n", vmcore_o->name);
		goto err;
	}
	Info("  vmcore bitmap      : written\n");

	Info("  skipped pages      : %d\n", skipped_page_cnt);
	Info("  saved pages        : %d\n", saved_page_cnt);

	Info("dump filtering completed.\n");

	exit(0);
err:

	if (unlink(vmcore_o->name) < 0) {
		Err("unable to delete %s\n", vmcore_o->name);
	}
	exit(1);
}
